@@ -247,18 +247,18 @@ gail_button_new (GtkWidget *widget)
 static G_CONST_RETURN gchar*
 gail_button_get_name (AtkObject *obj)
 {
+  G_CONST_RETURN gchar* name = NULL;
+
   g_return_val_if_fail (GAIL_IS_BUTTON (obj), NULL);
 
-  if (obj->name != NULL)
-    return obj->name;
-  else
+  name = ATK_OBJECT_CLASS (parent_class)->get_name (obj);
+  if (name == NULL)
     {
       /*
        * Get the text on the label
        */
       GtkWidget *widget;
       GtkWidget *child;
-      G_CONST_RETURN gchar* name = NULL;
 
       widget = GTK_ACCESSIBLE (obj)->widget;
       if (widget == NULL)
@@ -285,8 +285,8 @@ gail_button_get_name (AtkObject *obj)
               name = atk_object_get_name (atk_obj);
             }
         }
-      return name;
     }
+  return name;
 }
 
 static void
@@ -583,7 +583,8 @@ gail_button_get_keybinding (AtkAction *action,
          * We look for a mnemonic on the label
          */
         GtkWidget *widget;
-        GtkWidget *child;
+        GtkWidget *label;
+        guint key_val; 
 
         button = GAIL_BUTTON (action);
         widget = GTK_ACCESSIBLE (button)->widget;
@@ -595,17 +596,47 @@ gail_button_get_keybinding (AtkAction *action,
 
         g_return_val_if_fail (GTK_IS_BUTTON (widget), NULL);
 
-        child = get_label_from_button (widget, 0, FALSE);
-        if (GTK_IS_LABEL (child))
+        label = get_label_from_button (widget, 0, FALSE);
+        if (GTK_IS_LABEL (label))
           {
-            guint key_val; 
-
-            key_val = gtk_label_get_mnemonic_keyval (GTK_LABEL (child)); 
+            key_val = gtk_label_get_mnemonic_keyval (GTK_LABEL (label)); 
             if (key_val != GDK_VoidSymbol)
               return_value = gtk_accelerator_name (key_val, GDK_MOD1_MASK);
-            g_free (button->click_keybinding);
-            button->click_keybinding = return_value;
           }
+        if (return_value == NULL)
+          {
+            /* Find labelled-by relation */
+            AtkRelationSet *set;
+            AtkRelation *relation;
+            GPtrArray *target;
+            gpointer target_object;
+
+            set = atk_object_ref_relation_set (ATK_OBJECT (action));
+            if (set)
+              {
+                relation = atk_relation_set_get_relation_by_type (set, ATK_RELATION_LABELLED_BY);
+                if (relation)
+                  {              
+                    target = atk_relation_get_target (relation);
+            
+                    target_object = g_ptr_array_index (target, 0);
+                    if (GTK_IS_ACCESSIBLE (target_object))
+                      {
+                        label = GTK_ACCESSIBLE (target_object)->widget;
+                      } 
+                  }
+                g_object_unref (set);
+              }
+
+            if (GTK_IS_LABEL (label))
+              {
+                key_val = gtk_label_get_mnemonic_keyval (GTK_LABEL (label)); 
+                if (key_val != GDK_VoidSymbol)
+                  return_value = gtk_accelerator_name (key_val, GDK_MOD1_MASK);
+              }
+          }
+        g_free (button->click_keybinding);
+        button->click_keybinding = return_value;
         break;
       }
     default:
@@ -1342,6 +1373,82 @@ gail_button_finalize (GObject            *object)
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
+static GtkWidget*
+find_label_child (GtkContainer *container,
+                  gint         *index,
+                  gboolean     allow_many)
+{
+  GList *children, *tmp_list;
+  GtkWidget *child;
+ 
+  children = gtk_container_get_children (container);
+
+  child = NULL;
+  for (tmp_list = children; tmp_list != NULL; tmp_list = tmp_list->next) 
+    {
+      if (GTK_IS_LABEL (tmp_list->data))
+        {
+          if (!allow_many)
+            {
+              if (child)
+                {
+                  child = NULL;
+                  break;
+                }
+              child = GTK_WIDGET (tmp_list->data);
+            }
+          else
+            {
+              if (*index == 0)
+                {
+                  child = GTK_WIDGET (tmp_list->data);
+                  break;
+                }
+              *index--;
+	    }
+        }
+       /*
+        * Label for button which are GtkTreeView column headers are in a 
+        * GtkHBox in a GtkAlignment.
+        */
+      else if (GTK_IS_ALIGNMENT (tmp_list->data))
+        {
+          GtkWidget *widget;
+
+          widget = gtk_bin_get_child (GTK_BIN (tmp_list->data));
+          if (GTK_IS_LABEL (widget))
+            {
+              if (!allow_many)
+                {
+                  if (child)
+                    {
+                      child = NULL;
+                      break;
+                    }
+                  child = widget;
+                }
+              else
+                {
+                  if (*index == 0)
+                    {
+	              child = widget;
+                      break;
+                    }
+                  *index--;
+	        }
+	    }
+        }
+      else if (GTK_IS_CONTAINER (tmp_list->data))
+        {
+          child = find_label_child (GTK_CONTAINER (tmp_list->data), index, allow_many);
+          if (child)
+            break;
+        } 
+    }
+  g_list_free (children);
+  return child;
+}
+
 static GtkWidget*
 get_label_from_button (GtkWidget *button,
                        gint      index,
@@ -1358,78 +1465,48 @@ get_label_from_button (GtkWidget *button,
   if (GTK_IS_ALIGNMENT (child))
     child = gtk_bin_get_child (GTK_BIN (child));
 
-  while (GTK_IS_CONTAINER (child))
-    {
-      /*
-       * Child is not a label.  Search for a label in the container's children.
-       */
-      GList *children, *tmp_list;
- 
-      children = gtk_container_get_children (GTK_CONTAINER (child));
-
-      child = NULL;
-      for (tmp_list = children; tmp_list != NULL; tmp_list = tmp_list->next) 
-	{
-	  if (GTK_IS_LABEL (tmp_list->data))
-	    {
-              if (!allow_many)
-                {
-                  if (child)
-                    {
-                      child = NULL;
-                      break;
-                    }
-	          child = GTK_WIDGET (tmp_list->data);
-                }
-              else
-                {
-                  if (n_labels == index)
-                    {
-	              child = GTK_WIDGET (tmp_list->data);
-                      break;
-                    }
-                  n_labels++;
-	        }
-	    }
-           /*
-            * Label for button which are GtkTreeView column headers are in a 
-            * GtkHBox in a GtkAlignment.
-            */
-          if (GTK_IS_ALIGNMENT (tmp_list->data))
-            {
-              GtkWidget *widget;
-
-              widget = gtk_bin_get_child (GTK_BIN (tmp_list->data));
-              if (GTK_IS_LABEL (widget))
-                {
-                  if (!allow_many)
-                    {
-                      if (child)
-                        {
-                          child = NULL;
-                          break;
-                        }
-	              child = GTK_WIDGET (tmp_list->data);
-                    }
-                  else
-                    {
-                      if (n_labels == index)
-                        {
-	                  child = GTK_WIDGET (tmp_list->data);
-                          break;
-                        }
-                      n_labels++;
-	            }
-	        }
-	    }
-	}
-
-      g_list_free (children);
-    }
+  if (GTK_IS_CONTAINER (child))
+    child = find_label_child (GTK_CONTAINER (child), &index, allow_many);
+  else if (!GTK_IS_LABEL (child))
+    child = NULL;
 
   return child;
 }
 
+static void*
+count_labels (GtkContainer *container,
+              gint         *n_labels)
+{
+  GList *children, *tmp_list;
+ 
+  children = gtk_container_get_children (container);
+
+  for (tmp_list = children; tmp_list != NULL; tmp_list = tmp_list->next) 
+    {
+      if (GTK_IS_LABEL (tmp_list->data))
+        {
+          *n_labels++;
+        }
+       /*
+        * Label for button which are GtkTreeView column headers are in a 
+        * GtkHBox in a GtkAlignment.
+        */
+      else if (GTK_IS_ALIGNMENT (tmp_list->data))
+        {
+          GtkWidget *widget;
+
+          widget = gtk_bin_get_child (GTK_BIN (tmp_list->data));
+          if (GTK_IS_LABEL (widget))
+            *n_labels++;
+        }
+      else if (GTK_IS_CONTAINER (tmp_list->data))
+        {
+          count_labels (GTK_CONTAINER (tmp_list->data), n_labels);
+        } 
+    }
+  g_list_free (children);
+}
+
 static gint
 get_n_labels_from_button (GtkWidget *button)
 {
@@ -1441,35 +1518,10 @@ get_n_labels_from_button (GtkWidget *button)
   child = gtk_bin_get_child (GTK_BIN (button));
   if (GTK_IS_ALIGNMENT (child))
     child = gtk_bin_get_child (GTK_BIN (child));
-  while (GTK_IS_CONTAINER (child))
-    {
-      /*
-       * Child is not a label.  Search for a label in the container's children.
-       */
-      GList *children, *tmp_list;
- 
-      children = gtk_container_get_children (GTK_CONTAINER (child));
 
-      child = NULL;
-      for (tmp_list = children; tmp_list != NULL; tmp_list = tmp_list->next) 
-	{
-	  if (GTK_IS_LABEL (tmp_list->data))
-            n_labels++;
+  if (GTK_IS_CONTAINER (child))
+    count_labels (GTK_CONTAINER (child), &n_labels);
 
-           /*
-            * Label for button which are GtkreeView column headers are in a 
-            * GtkHBox in a GtkAlignment.
-            */
-          if (GTK_IS_ALIGNMENT (tmp_list->data))
-            {
-              child = gtk_bin_get_child (GTK_BIN (tmp_list->data));
-	      if (GTK_IS_LABEL (child))
-                n_labels++;
-	    }
-	}
-
-      g_list_free (children);
-    }
   return n_labels;
 }
 
