@@ -318,127 +318,47 @@ gail_widget_get_parent (AtkObject *accessible)
 }
 
 static GtkWidget*
-find_label (GtkWidget *start,
-            GtkWidget *widget,
-            gboolean  search_all)
+find_label (GtkWidget *widget)
 {
-
-  GList *list, *l;
-  GtkWidget *parent;
+  GList *labels;
   GtkWidget *label;
+  GtkWidget *temp_widget;
 
+  labels = gtk_widget_list_mnemonic_labels (widget);
   label = NULL;
-  parent = gtk_widget_get_parent (start);
-  if (GTK_IS_CONTAINER (parent))
+  if (labels)
     {
-      list = gtk_container_get_children (GTK_CONTAINER (parent));
-
-      l = list;
-      if (l)
+      if (labels->data)
         {
-          while (l)
+          if (labels->next)
             {
-              if ((GtkWidget *)l->data == start)
-                break;
-
-              l = l->next;
+              g_warning ("Widget (%s) has more than one label", G_OBJECT_TYPE_NAME (widget));
+              
             }
-          /*
-           * The widget may claim container as parent but it may not
-           * be listed as a child.
-           */
-          if (!l)
+          else
             {
-              if (GTK_IS_SCROLLED_WINDOW (parent) && GTK_IS_SCROLLBAR (start))
-                return NULL;
+              label = labels->data;
             }
-          g_return_val_if_fail (l, NULL);
-          if (search_all)
-            {
-              GList *temp_l;
-
-              temp_l = l;
-              while (temp_l->prev)
-                {
-                  if (GTK_IS_LABEL (temp_l->prev->data))
-	            {  
-                      label = GTK_WIDGET (temp_l->prev->data);
-                      if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
-                        label = NULL;
-                      else
-                        break;
-                    }
-                  /* Heuristic to find label in GtkFileSelection */
-                  else if (GTK_IS_EVENT_BOX (temp_l->prev->data))
-                    {
-                       label = GTK_BIN (temp_l->prev->data)->child;
-                       if (GTK_IS_LABEL (label) &&
-                           gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
-                         label = NULL;
-                       else
-                         break;
-                    }
-                  temp_l = temp_l->prev;
-                }
-              if (label == NULL)
-                {
-                  temp_l = l;
-                  while (temp_l->next)
-                    {
-                      if (GTK_IS_LABEL (temp_l->next->data))
-	                {  
-                          label = GTK_WIDGET (temp_l->next->data);
-                          if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
-                            label = NULL;
-                          else
-                            break;
-                        }
-                      /* Heuristic to find label in NautilusFilePropertiesWindow
-                         bug #134629 */
-                      if (GTK_IS_HBOX (temp_l->next->data))
-                        {
-                          GList *box_children, *temp_children;
-
-                          temp_children = box_children = gtk_container_get_children (GTK_CONTAINER (temp_l->next->data));
-                          while (temp_children)
-                            {
-                              if (GTK_IS_LABEL (temp_children->data))
-                                {
-                                  label = GTK_WIDGET (temp_children->data);
-                                  if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
-                                    label = NULL;
-                                  else
-                                    break;
-                                }
-                              temp_children = temp_children->next;
-                            }
-                          g_list_free (box_children);
-                          if (label)
-                            break;
-                        }
-                      temp_l = temp_l->next;
-                    }
-                }
-            }
-          else 
-            { 
-              if (l->prev && GTK_IS_LABEL (l->prev->data))
-	        {  
-                  label = GTK_WIDGET (l->prev->data);
-                  if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
-                    label = NULL;
-                }
-              if (!label && l->next && GTK_IS_LABEL (l->next->data))
-	        {  
-                  label = GTK_WIDGET (l->next->data);
-                  if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
-                    label = NULL;
-                }
-            } 
-
-          g_list_free (list);
-       }
+        }
+      g_list_free (labels);
     }
+
+  /*
+   * Ignore a label within a button; bug #136602
+   */
+  if (label && GTK_IS_BUTTON (widget))
+    {
+      temp_widget = label;
+      while (temp_widget)
+        {
+          if (temp_widget == widget)
+            {
+              label = NULL;
+              break;
+            }
+          temp_widget = gtk_widget_get_parent (temp_widget);
+        }
+    } 
   return label;
 }
 
@@ -471,93 +391,27 @@ gail_widget_ref_relation_set (AtkObject *obj)
 
   if (!atk_relation_set_contains (relation_set, ATK_RELATION_LABELLED_BY))
     {
-      /*
-       * Heuristic to create LABELLED_BY relation:
-       *
-       * If any sibling the current widget is a label 
-       * and its mnemonic widget is the current widget then set LABELLED_BY 
-       * relation on current accessible object.
-       *
-       * If this fails try to find a label at the same level as the widget's
-       * parent. This finds labels for NautilusProgramChooser dialog.
-       */
-      label = find_label (widget, widget, TRUE);
-      if (label == NULL)
+      label = find_label (widget);
+      if (label == NULL && GTK_IS_BUTTON (widget))
+        /*
+         * Handle the case where GnomeIconEntry is the mnemonic widget.
+         * The GtkButton which is a grandchild of the GnomeIconEntry
+         * should really be the mnemonic widget. See bug #133967.
+         */
         {
           GtkWidget *temp_widget;
 
           temp_widget = gtk_widget_get_parent (widget);
-          if (temp_widget)
-            label = find_label (temp_widget, widget, FALSE);
-          /*
-           * This finds relationship in Desktop Preferences Background.
-           * See bug #126973.
-           * This finds relationship for GnomeEntry inside a GnomeFileEntry
-	   * and its label.
-           * See bug #133403.
-           */
-          if (!label)
+
+          if (GTK_IS_ALIGNMENT (temp_widget))
             {
-              if ((GTK_IS_VBOX (temp_widget) && GTK_IS_OPTION_MENU (widget)) ||
-                  (GTK_IS_HBOX (temp_widget) && GTK_IS_COMBO (widget)))
+              temp_widget = gtk_widget_get_parent (temp_widget);
+              if (GTK_IS_BOX (temp_widget))
                 {
-                  temp_widget = gtk_widget_get_parent (temp_widget);
-                  if (temp_widget)
-                    {
-                      if (GTK_IS_OPTION_MENU (widget))
-                        label = find_label (temp_widget, widget, FALSE);
-                      else if (GTK_IS_COMBO (widget))
-                        label = find_label (temp_widget, temp_widget, FALSE);
-                      if (!label)
-                        {
-                          /*
-                           * Find label for combo-box in Sound preferences
-                           * capplet. See bug #126972.
-                           */
-                          temp_widget = gtk_widget_get_parent (temp_widget);
-                          if (GTK_IS_VBOX (temp_widget))
-                            {
-                              GList *list;
-
-                              list = gtk_container_get_children (GTK_CONTAINER (temp_widget));
-                              if (list)
-                                {
-                                  temp_widget = list->data;
-                                  if (GTK_IS_VBOX (temp_widget))
-                                    {
-                                      g_list_free (list);
-                                      list = gtk_container_get_children (GTK_CONTAINER (temp_widget));
-                                      if (list)
-                                        {
-                                          temp_widget = list->data;
-                                          g_list_free (list);
-                                          if (GTK_IS_LABEL (temp_widget) &&
-                                              gtk_label_get_mnemonic_widget (GTK_LABEL (temp_widget)) == widget)
-                                            {
-                                              label = temp_widget; 
-                                            }
-                                        }
-                                    }
-                                }
-                            }
-
-                        }
-                    }
-                }
-              /*
-               * Handle the case where GnomneIconEntry is the mnemonic widget.
-               * The GtkButton which is a grandchild of the GnomeIconEntry
-               * should really be the mnemonic widget. See bug #133967.
-               */
-              else if (GTK_IS_BUTTON (widget) && GTK_IS_ALIGNMENT (temp_widget))
-                {
-                  temp_widget = gtk_widget_get_parent (temp_widget);
-                  if (GTK_IS_BOX (temp_widget))
-                    {
-                      label = find_label (temp_widget, temp_widget, TRUE);
-                      if (!label)
-                        label = find_label (gtk_widget_get_parent (temp_widget), temp_widget, FALSE);
-                    }
+                  label = find_label (temp_widget);
+                 
+                  if (!label)
+                    label = find_label (gtk_widget_get_parent (temp_widget));
                 }
             }
         }
